"""
Module for interacting with RP1210 adapter drivers, following the RP1210C standard.
Contributions welcome!

Author: Darius Fieschko

While I try to provide adequate documentation, the RP1210C standard is owned by TMC, not me, and is
not reproduced here. For a complete understanding of the RP1210 standard, you must purchase and
read the official RP1210C documentation from TMC.

Official RP1210C documentation can be purchased from TMC at this link ($37.50 at time of writing):
    https://www.atabusinesssolutions.com/Shopping/Product/viewproduct/2675472/TMC-Individual-RP
"""

def sanitize_msg_param(param, num_bytes : int = 0, byteorder : str = 'big') -> bytes:
    """
    'Sanitizes' (converts to bytes) a message parameter.

    Defaults to big-endianness and whatever the size of param is.

    This function is meant for internal use in message/protocol files; it's only public because
    I didn't want to copy/paste it a bunch of times.
    """
    if isinstance(param, int): # int to bytes
        if num_bytes == 0:
            num_bytes = (param.bit_length() + 7) // 8
            if param == 0: # don't cut it off if the input is zero
                num_bytes = 1
        return param.to_bytes(num_bytes, byteorder)
    elif isinstance(param, str): # string to bytes
        if param == "": # check for empty string
            return b'' + b'\x00' * num_bytes
        return sanitize_msg_param(str.encode(param, 'utf8'), num_bytes, byteorder)
    elif isinstance(param, bytes):
         # convert to int, run sanitize_msg_param again
        if num_bytes == 0:
            if param == b'': # len == 1 for b'', but we don't want to return b'\x00'
                return b''
            num_bytes = len(param)
        if byteorder == 'little':
            param2 = param[::-1]
        else:
            param2 = param
        val = int.from_bytes(param2[:num_bytes], byteorder)
        return sanitize_msg_param(val, num_bytes, byteorder)
    else:
        try:
            return sanitize_msg_param(bytes(param), num_bytes, byteorder)
        except Exception:
            raise TypeError('sanitize_msg_param() param must be int, bool, str, or bytes', param)
    

# Import everything from RP1210.py
from RP1210.RP1210 import *
# Import other modules (not necessary in Python 3.9+)
from RP1210 import Commands, J1939



"""
Generate ClientCommand inputs for RP1210_SendCommand.

Each function in this file returns a value that can be used for ClientCommand.
"""
from typing import Literal
from RP1210 import sanitize_msg_param

COMMAND_IDS = {
    "RESET_DEVICE" : 0,
    "SET_ALL_FILTERS_STATES_TO_PASS" : 3,
    "SET_MESSAGE_FILTERING_FOR_J1939" : 4,
    "SET_MESSAGE_FILTERING_FOR_CAN" : 5,
    "SET_MESSAGE_FILTERING_FOR_J1708" : 7,
    "SET_MESSAGE_FILTERING_FOR_J1850" : 8,
    "SET_MESSAGE_FILTERING_FOR_ISO15765" : 9,
    "GENERIC_DRIVER_COMMAND" : 14,
    "SET_J1708_MODE" : 15,
    "ECHO_TRANSMITTED_MESSAGES" : 16,
    "SET_ALL_FILTERS_STATES_TO_DISCARD" : 17,
    "SET_MESSAGE_RECEIVE" : 18,
    "PROTECT_J1939_ADDRESS" : 19,
    "SET_BROADCAST_FOR_J1708" : 20,
    "SET_BROADCAST_FOR_CAN" : 21,
    "SET_BROADCAST_FOR_J1939" : 22,
    "SET_BROADCAST_FOR_J1850" : 23,
    "SET_J1708_FILTER_TYPE" : 24,
    "SET_J1939_FILTER_TYPE" : 25,
    "SET_CAN_FILTER_TYPE" : 26,
    "SET_J1939_INTERPACKET_TIME" : 27,
    "SET_MAX_ERROR_MSG_SIZE" : 28,
    "DISALLOW_FURTHER_CONNECTIONS" : 29,
    "SET_J1850_FILTER_TYPE" : 30,
    "RELEASE_J1939_ADDRESS" : 31,
    "SET_ISO15765_FILTER_TYPE" : 32,
    "SET_BROADCAST_FOR_ISO15765" : 33,
    "SET_ISO15765_FLOW_CONTROL" : 34,
    "CLEAR_ISO15765_FLOW_CONTROL" : 35,
    "SET_J1939_BAUD" : 37,
    "SET_ISO15765_BAUD" : 38,
    "SET_BLOCKTIMEOUT" : 215,
    "SET_J1708_BAUD" : 305,
    "FLUSH_TX_RX_BUFFERS" : 39,
    "SET_BROADCAST_FOR_KWP2000" : 41,
    "SET_BROADCAST_FOR_ISO9141" : 42,
    "GET_PROTOCOL_CONNECTION_SPEED" : 45,
    "SET_ISO9141KWP2000_MODE" : 46,
    "SET_CAN_BAUD" : 47,
    "GET_WIRELESS_STATE" : 48}
"""
Mnemonics for RP1210_SendCommand commands. Follows ordering of table in section 21.4.

key:item orders are swapped from RP1210_COMMANDS in RP1210.py.

Example:
    command_id = Commands.COMMAND_IDS["PROTECT_J1939_ADDRESS"]
"""

J1939_FILTERS = {
    "PGN" : 0x01,
    "SOURCE" : 0x04,
    "DEST" : 0x08,
    "DESTINATION" : 0x08
}
"""
A dict of valid J1939 filter parameters.

When using these as an argument for RP1210_Set_Message_Filtering_For_J1939,
you can add them together to filter for more than one at once.

Example (this will set the filter for both source and destination addresses):
    filter_flag = J1939_FILTERS["SOURCE"] + J1939_FILTERS["DEST"]

- "PGN"         => 1
- "SOURCE"      => 4
- "DEST"        => 8
- "DESTINATION" => 8
"""

CAN_TYPES = {
    "STANDARD_CAN" : 0x00,
    "EXTENDED_CAN" : 0x01,
    "STANDARD_CAN_IS015765_EXTENDED" : 0x02,
    "EXTENDED_CAN_IS015765_EXTENDED" : 0x03,
    "STANDARD_MIXED_CAN_IS015765" : 0x04
}
"""
A dict of CAN type codes, as defined in the RP1210C standard.
- "STANDARD_CAN" : 0x00
- "EXTENDED_CAN" : 0x01
- "STANDARD_CAN_IS015765_EXTENDED" : 0x02
- "EXTENDED_CAN_IS015765_EXTENDED" : 0x03
- "STANDARD_MIXED_CAN_IS015765" : 0x04
"""

ECHO_MODES = {
    "ECHO_OFF" : 0x00,
    "ECHO_ON" : 0x01
}
"""
Dict for echo modes, used in commmand ECHO_TRANSMITTED_MESSAGES (echoTx())

- ECHO_OFF : 0x00
- ECHO_ON : 0x01
"""

RECEIVE_MODES = {
    "RECEIVE_ON" : 0x01,
    "RECEIVE_OFF" : 0x00
}
"""
Dict for receive modes, used in command SET_MESSAGE_RECEIVE
- RECEIVE_ON : 0x01
- RECEIVE_ OFF : 0x00
"""

BROADCAST_FUNCTIONS = {
    "ADD_LIST" : 1,
    "VIEW_LIST" : 2,
    "DESTROY_LIST" : 3,
    "REMOVE_ENTRY" : 4,
    "LIST_LENGTH" : 5
}
"""
Dict for different broadcast functions.

- ADD_LIST : 1
- VIEW_LIST : 2
- DESTROY_LIST : 3
- REMOVE_ENTRY : 4
- LIST_LENGTH : 5
"""

BAUD_RATES = {

}

def resetDevice():
    """
    Reset Device (0) (0 bytes)

    RP1210_RESET_DEVICE only works if only one client is connected to the adapter, and does
    the exact same thing as if you called the function ClientDisconnect.

    Returns empty byte string (this command needs no extra data).
    """
    return b''

def setAllFiltersToPass():
    """
    Set All Filter States to Pass (3) (0 bytes)

    Returns empty byte string (this command needs no extra data).
    """
    return b''

def setJ1939Filters(filter_flag : int, pgn = 0, source = 0, dest = 0) -> bytes:
    """
    Set Message Filtering for J1939 (4) (7 bytes)

    Args:
    - Filter flag (1 byte) - filter flag integer/bytecode.
        - FILTER_PGN (1), FILTER_SOURCE (2), FILTER_DESTINATION (4)
        - J1939_FILTERS dict is available for convenience.
    - PGN (3 bytes) - the PGN that needs to be filtered.
    - Source Address (1 byte) - the source address that needs to be filtered.
    - Destination Address (1 byte) - the destination address that needs to be filtered.

    Args pgn, source, and dest only do anything if they are set with filter_flag. If they aren't,
    they will be ignored.

    You can specifiy filter_flag and keyword arguments instead of entering useless values for
    pgn, source, or dest.
        Example (this will filter for messages that come from 0x1E and sent to 0xB0):
            command = Commands.setJ1939Filters(4+8, source=0x1E, dest=0xB0)
    """
    ret_val = sanitize_msg_param(filter_flag, 1)
    ret_val += sanitize_msg_param(pgn, 3, 'little')
    ret_val += sanitize_msg_param(0, 1) # FILTER_PRIORITY was removed from RP1210 standard
    ret_val += sanitize_msg_param(source, 1)
    ret_val += sanitize_msg_param(dest, 1)
    return ret_val 


def setCANFilters(can_type, mask, header) -> bytes:
    """
    Set Message Filtering for CAN (5) (9 bytes)

    Args:
    - CAN Type (1 byte) - 0x00 for STANDARD_CAN, 0x01 for EXTENDED_CAN.
        - See dict CAN_TYPES for other types.
    - Mask (4 bytes) - a bitwise mask that indicates which bits in the header need to be matched.
        - Big endian; "1" means a value is important; "0" means a value is unimportant.
    - Header (4 bytes) - "Indicates what value is required for each bit of interest".

    This is one of those functions that you're going to want the RP1210C documentation for.
    """
    ret_val = sanitize_msg_param(can_type, 1)
    ret_val += sanitize_msg_param(mask, 4)
    ret_val += sanitize_msg_param(header, 4)
    return ret_val

def generic(ClientCommand, num_bytes = 0, byteorder = 'big') -> bytes:
    """
    Generic Driver Command (14)

    Args:
    - ClientCommand - the bytes/buffer to be sent to the drivers.
    - size = size of ClientCommand. Leave 0 to automatically size to len(ClientCommand).
    - byteorder = big endian ('big') or little endian ('little')
    """
    return sanitize_msg_param(ClientCommand, num_bytes, byteorder)

def setEcho(echo = True) -> bytes:
    """
    Set Echo Transmitted Messages (16) (1 byte)

    Args:
    - Echo on/off (bool) - False for no echo, True for echo.
    """
    if echo:
        msg = ECHO_MODES["ECHO_ON"]
    else:
        msg = ECHO_MODES["ECHO_OFF"]
    return sanitize_msg_param(msg, 1)

def setAllFiltersToDiscard():
    """
    Set All Filter States to Discard (17) (0 bytes)

    Returns empty byte string (this command needs no extra data).
    """
    return b''

def setMessageReceive(receive_messages = True) -> bytes:
    """
    Set Message Receive (18) (1 byte)

    Args:
    - Receive on/off : True = RECEIVE_ON, False = RECEIVE_OFF.
    """
    if receive_messages:
        msg = RECEIVE_MODES["RECEIVE_ON"]
    else:
        msg = RECEIVE_MODES["RECEIVE_OFF"]
    return sanitize_msg_param(msg, 1)

def protectJ1939Address(address_to_claim, network_mgt_name, blocking = True) -> bytes:
    """
    Protect J1939 Address (19) (10 bytes)

    This command claims an address on the J1939 bus.
    - address_to_claim (1 byte) - 8-bit address to claim on the J1939 bus.
    - network_mgt_name (8 bytes) - 8-byte name of client on network (this is you!)
        - See J1939 network management standard!
        - Lowest name takes priority if two devices try to claim the same address
    - blocking (bool) - True will block until done, False will return before completion

    This function automatically sanitizes str, int, and bytes inputs. str are parsed as 10-bit
    decimals! Use byte strings (b"message") if you want to pass utf-8 characters.
    """
    addr = sanitize_msg_param(address_to_claim, 1)
    name = sanitize_msg_param(network_mgt_name, 8)
    if blocking:
        status = sanitize_msg_param(0, 1) # BLOCK_UNTIL_DONE
    else:
        status = sanitize_msg_param(2, 1) # RETURN_BEFORE_COMPLETION
    return addr + name + status

def releaseJ1939Address(address) -> bytes:
    """
    Release a J1939 Address (31) (1 byte)

    Args:
    - Address (1 byte) - the address to release.

    This doesn't do anything special with the J1939 bus. All it does is tell your adapter not to
    use this address anymore.
    """
    return sanitize_msg_param(address, 1)
    

def setBroadcastList(function : Literal[1, 2, 3, 4, 5], command) -> bytes:
    """
    A catch-all function for each RP1210_Set_Broadcast_For_XXX command, since each one takes
    the same type of ClientCommand argument.

    Function codes:
    - 1 = ADD_LIST
    - 2 = VIEW_B_LIST
    - 3 = DESTROY_LIST
    - 4 = REMOVE_ENTRY
    - 5 = LIST_LENGTH

    Param command is equivalent to fpchClientCommand[1..n] input to RP1210_SendCommand.

    This function is mostly useful for building the following functions:
    - addBroadcastList
    - viewBroadcastList
    - destroyBroadcastList
    - removeBroadcastEntry
    - getBroadcastListLength

    This function applies for protocols J1708, CAN, J1939, J1850, ISO15765, ISO9141, and KWP2000.
    """
    func = sanitize_msg_param(function, 1)
    # if not func in [b'\x01', b'\x02', b'\x03', b'\x04', b'\x05']:
    #     raise ValueError("function must be one of:", "ADD_LIST (1)", "VIEW_B_LIST (2)", 
    #                         "DESTROY_LIST (3)", "REMOVE_ENTRY (4)", "LIST_LENGTH (5")
    ret_val = func
    ret_val += sanitize_msg_param(command)
    return ret_val


def setFilterType(filter_type : Literal[0, 1]):
    """
    Set Filter Type (1 byte)

    filter_type:
    - 0 = FILTER_INCLUSIVE
    - 1 = FILTER_EXCLUSIVE

    Used for:
    - RP1210_Set_J1708_Filter_Type (24)
    - RP1210_Set_J1939_Filter_Type (25)
    - RP1210_Set_CAN_Filter_Type (26)
    - RP1210_Set_J1850_Filter_Type (30)
    - RP1210_Set_IS015765_Filter_Type (32)
    """
    val = sanitize_msg_param(filter_type, 1)
    # if not val in [b'\x00', b'\x01']:
    #     raise ValueError("filter_type must be 0 (FILTER_INCLUSIVE) or 1 (FILTER_EXCLUSIVE)")
    return val

def setJ1939InterpacketTime(time_in_ms : int):
    """
    Set J1939 Broadcast Interpacket Timing (27)

    Args:
    - time_in_ms - interpacket time in milliseconds (unsigned 32-bit int)
    """
    return sanitize_msg_param(time_in_ms, 4, 'little')

def setMaxErrorMsgSize(msg_size : int):
    """
    Set Max Error Message Return Size (28)

    Args:
    - msg_size - value in bytes for how large error messages are allowed to be.
        - Should be between 81 and 65535
    """
    return sanitize_msg_param(msg_size, 2, 'little')

def disallowConnections():
    """
    Disallow Further Client Connections (29)

    There is no data required for this command, so this function will return b''
    """
    return b''

def setJ1939Baud(baud_code : int, wait_for_msg = True):
    """
    Set J1939 Baud Rate (37)

    Args:
    - baud_code - code that corresponds w/ desired baud rate
        - 125k = 4
        - 250k = 5
        - 500k = 6
        - 1000k = 7
    - wait_for_msg - should we apply the baud change after the current message is finished, or
                    apply the change right away?
    """
    # translate codes if applicable
    if baud_code in [125000, 125, '125', '125k', '125000']:
        baud_code = 0x04
    elif baud_code in [250000, 250, '250', '250k', '250000']:
        baud_code = 0x05
    elif baud_code in [500000, 500, '500', '500k', '500000']:
        baud_code = 0x06
    elif baud_code in [1000000, 1000, '1000', '1000000', '1000k']:
        baud_code = 0x07
    # check wait_for_msg flag
    if wait_for_msg:
        wait_msg = b'\x01'
    else:
        wait_msg = b'\x00'
    # return value
    return wait_msg + sanitize_msg_param(baud_code, 1)


def setBlockingTimeout(block1 : int, block2 : int):
    """
    Set Blocking Timeout (215) (2 bytes)

    Block 1 and block 2 are multiplied together to determine the final blocking time in
    milliseconds. Set either block to 0 for infinite time.
    """
    return sanitize_msg_param(block1, 1) + sanitize_msg_param(block2, 1)

def flushBuffers():
    """
    Flush the Send/Receive Buffers (39)

    There is no data required for this command, so this function will return b''
    """
    return b''

def getConnectionSpeed():
    """
    What Speed Did the VDA Connect? (45)

    There is no data required for this command, so this function will return b''
    """
    return b''

def setCANBaud(baud_code : int, wait_for_msg = True):
    """
    Set CAN Baud Rate (47)

    Args:
    - baud_code - code that corresponds w/ desired baud rate
        - 9600 = 0
        - 19200 = 1
        - 38400 = 2
        - 57600 = 3
        - 125k = 4
        - 250k = 5
        - 500k = 6
        - 1000k = 7
    - wait_for_msg - should we apply the baud change after the current message is finished, or
                    apply the change right away?
    """
    # translate codes if applicable
    if baud_code in [9600, 9.6, '9600', '9.6k']:
        baud_code = 0x00
    elif baud_code in [19200, 19.2, '19200', '19.2k']:
        baud_code = 0x01
    elif baud_code in [38400, 38.4, '38400', '38.4k']:
        baud_code = 0x02
    elif baud_code in [57600, 57.6, '57600', '57.6k']:
        baud_code = 0x03
    elif baud_code in [125000, 125, '125', '125k', '125000']:
        baud_code = 0x04
    elif baud_code in [250000, 250, '250', '250k', '250000']:
        baud_code = 0x05
    elif baud_code in [500000, 500, '500', '500k', '500000']:
        baud_code = 0x06
    elif baud_code in [1000000, 1000, '1000', '1000000', '1000k']:
        baud_code = 0x07
    # check wait_for_msg flag
    if wait_for_msg:
        wait_msg = b'\x01'
    else:
        wait_msg = b'\x00'
    # return value
    return wait_msg + sanitize_msg_param(baud_code, 1)



"""
J1939 functions for use with RP1210 APIs. Note that this implementation is RP1210-specific - i.e.
if you try to generate J1939 messages to transmit directly onto a CANbus, these functions will fail!

While a dict of J1939 PGNs would be convenient, they are not provided here because the list is a
copyright of SAE.
"""

from RP1210 import sanitize_msg_param

def toJ1939Message(pgn, priority, source, destination, data) -> bytes:
    """
    Converts args to J1939 message suitable for RP1210_SendMessage function.

    String arguments are read as base-10! If you want to provide an argument in base-16, do it as
    an int or byte string.

    RP1210_SendMessage J1939 format:
    - PGN (3 bytes)
    - Priority (1 byte)
    - Source Address (1 byte)
    - Destination Address (1 byte)
    - Message Data (0 - 1785 byes)

    If you're not sure what to do with some of these arguments due to differences between PDU1 and
    PDU2 messages, you can most likely get away with leaving irrelevant portions blank - your RP1210
    adapter drivers will format the message for you.

    Arguments can be strings, ints, or bytes. This function will parse strings as UTF-8 characters,
    so don't provide it with letters or special characters unless that's what you mean to send.
    If you want to send it 0xFF, send it as an int and not "FF". Likewise, 0 != "0".
    """
    ret_val = sanitize_msg_param(pgn, 3, 'little')
    ret_val += sanitize_msg_param(priority, 1)
    ret_val += sanitize_msg_param(source, 1)
    ret_val += sanitize_msg_param(destination, 1)
    ret_val += sanitize_msg_param(data)
    return ret_val

def toJ1939Request(pgn_requested, source, destination = 255, priority = 6) -> bytes:
    """
    Formats a J1939 request message. This puts out a request for the specified PGN, and will prompt
    other devices on the network to respond.

    - pgn_requested: the parameter group you're requesting
    - source: the source address, e.g. your address on the CANBus
    - destination: 255 = global request; enter a different number to request from a specific address
    - priority: priority of request; default is 6
    """
    pgn_request = sanitize_msg_param(pgn_requested, 3, 'little') # must be little-endian
    return toJ1939Message(0x00EA00, priority, source, destination, pgn_request)

class DTC():
    """
    A convenience class for parsing or generating the diagnostic trouble code (DTC) in diagnostic
    message data. This class should work for DM1, DM2, and DM12 messages.
    
    To read SPN, FMI, and OC from a DTC, initialize this class with four bytes containing the DTC
    you want to parse: `dtc = DTC(dtc=data)`

    To generate a DTC, initialize this class with spn, fmi, and oc:
    `dtc = DTC(spn=532, fmi=4, oc=7)`

    It is also possible to generate an empty object `dtc = DTC()` and set its properties individually.

    Properties `data`, `spn`, `fmi`, and `oc` are all intelligently handled with setters. This means you
    can e.g. set `dtc.spn = 321`, and `dtc.data` will be updated accordingly.
    """

    def __init__(self, dtc : bytes = None, spn = 0, fmi = 0, oc = 0) -> None:
        self._spn = spn
        self._fmi = fmi
        self._oc = oc
        if dtc is None:
            self.data = self.to_bytes(spn, fmi, oc)
        else:
            self.data = sanitize_msg_param(dtc, 4)

    ##############
    # PROPERTIES #
    ##############
    #region properties

    @property
    def spn(self):
        """SPN (suspect parameter number)"""
        return self.get_spn(self.data)

    @spn.setter
    def spn(self, val : int):
        """SPN (suspect parameter number)"""
        self.data = self.to_bytes(val, self.fmi, self.oc)

    @property
    def fmi(self):
        """FMI (failure mode identifier)"""
        return self.get_fmi(self.data)

    @fmi.setter
    def fmi(self, val : int):
        """FMI (failure mode identifier)"""
        self.data = self.to_bytes(self.spn, val, self.oc)

    @property
    def oc(self):
        """OC (occurrence count)"""
        return self.get_oc(self.data)

    @oc.setter
    def oc(self, val : int):
        """OC (occurrence count)"""
        self.data = self.to_bytes(self.spn, self.fmi, val)

    #endregion

    ##################
    # DUNDER METHODS #
    ##################
    #region dundermethods

    def __iadd__(self, val : int):
        """Overload += so OC can be incremented directly."""
        oc = min(self.oc + val, 126) # limit to 126
        self.oc = max(oc, 0) # shouldn't be lower than 0 either
        return self

    def __str__(self) -> str:
        return str(self.data)

    def __bytes__(self) -> bytes:
        return self.data

    def __int__(self) -> int:
        return int.from_bytes(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def __eq__(self, other) -> bool:
        try:
            return self.data == sanitize_msg_param(other, 4)
        except Exception:
            return False

    def __bool__(self) -> bool:
        return self.data != b'\x00\x00\x00\x00' and len(self.data) == 4

    #endregion

    ########################
    # STATIC/CLASS METHODS #
    ########################
    #region staticmethods

    @staticmethod
    def get_spn(dtc : bytes) -> int:
        """Parses and returns SPN from 4-byte DTC."""
        data = sanitize_msg_param(dtc, 4)
        # need to construct SPN from bytes 0-2
        ret_val = (data[2] >> 5 & 0b00000111) << 16
        ret_val += data[1] << 8
        ret_val += data[0]
        return ret_val

    @staticmethod
    def get_fmi(dtc : bytes) -> int:
        """Parses and returns FMI from 4-byte DTC."""
        data = sanitize_msg_param(dtc, 4)
        # parse FMI from byte 2
        return data[2] & 0b00011111 # return bits 0-4 of byte 2

    @staticmethod
    def get_oc(dtc : bytes) -> int:
        """Parses and returns OC from 4-byte DTC."""
        data = sanitize_msg_param(dtc, 4)
        # parse FMI from byte 3
        return int(data[3]) & 0b01111111 # return bits 0-6 of byte 5

    @staticmethod
    def get_cm(dtc : bytes) -> int:
        """Parses and returns CM from 4-byte DTC."""
        dtc = sanitize_msg_param(dtc, 4)
        # parse CM from byte 3
        return dtc[3] >> 7 & 1

    @staticmethod
    def to_bytes(spn : int, fmi : int, oc : int):
        """Generates 4-byte DTC from SPN, FMI, and OC."""
        ret_val = b''
        # bytes 0 and 1 are just SPN in little-endian format
        spn_bytes = sanitize_msg_param(spn, 3, 'little')
        ret_val += int.to_bytes(spn_bytes[0], 1, 'big')
        ret_val += int.to_bytes(spn_bytes[1], 1, 'big')
        # byte 2 is mix of SPN (3 bits) and fmi (5 bits)
        # we will handle this by doing bitwise operations on an int, then convert back to bytes
        byte4_int = (int(spn_bytes[2]) << 5) & 0b11100000
        byte4_int |= int(fmi) & 0b00011111
        ret_val += sanitize_msg_param(byte4_int)
        # byte 3 is mix of CM (1 bit) and OC (7 bits) - CM always set to 0
        ret_val += sanitize_msg_param(int(oc) & 0b01111111)
        return ret_val

    @staticmethod
    def to_int(spn : int, fmi : int, oc : int):
        """Generates 4-byte DTC from SPN, FMI, and OC as int."""
        ret_val = spn & 0b11111111
        ret_val = (ret_val << 8) + (spn >> 8 & 0b11111111)
        ret_val = (ret_val << 8) + ((spn >> 11 & 0b11100000) | (fmi & 0b00011111))
        ret_val = (ret_val << 8) + (oc & 0b01111111)
        return ret_val
    #endregion

class J1939Message():
    """
    A convenience class for parsing a J1939 message received from RP1210_ReadMessage.

    Message must include timestamp (4 bytes) in its leading bytes (this conforms w/ return value
    of RP1210_ReadMessage).

    - Initialize the object with the message received from ReadMessage.
    - If you used the command 'Set Echo Transmitted Messages' to turn echo on, set arg echo = True.
    """
    def __init__(self, j1939_message : bytes, echo = False) -> None:
        self.msg = j1939_message
        self.echo_offset = int(echo)

    ##############
    # PROPERTIES #
    ##############

    def getTimestamp(self) -> int:
        """Returns timestamp (4 bytes) as int."""
        return int.from_bytes(self.msg[0:4], 'big')

    def getPGN(self) -> int:
        """Returns PGN (3 bytes) as int."""
        start = 4 + self.echo_offset
        end = 6 + self.echo_offset
        return int.from_bytes(self.msg[start:end+1], 'little')

    def getPriority(self) -> int:
        """Returns Priority (1 byte) as int."""
        loc = 7 + self.echo_offset
        return int(self.msg[loc])

    def getSource(self) -> int:
        """Returns Source Address (1 byte) as int."""
        loc = 8 + self.echo_offset
        return int(self.msg[loc])

    def getSourceAddress(self) -> int:
        """Returns Source Address (1 byte) as int."""
        return int(self.getSource())

    def getDestination(self) -> int:
        """Returns Destination Address (1 byte) as int."""
        loc = 9 + self.echo_offset
        return int(self.msg[loc])

    def getData(self) -> bytes:
        """Returns message data (0 - 1785 bytes) as bytes."""
        loc = 10 + self.echo_offset
        return self.msg[loc:]

    def isEcho(self) -> bool:
        """Returns True if the message is an echo of a message you transmitted, False if not."""
        if self.echo_offset == 1:
            return int.from_bytes(self.msg[4]) == 0x01
        return False

    def isRequest(self) -> bool:
        """Returns true if PGN matches J1939 Request PGN."""
        return self.getPGN() == 0xEA00

    def isDMRequest(self) -> bool:
        """Returns true if PGN matches Diagnostic Message Request PGN."""
        return self.getPGN() == 0xEA00

    def isDM1(self) -> bool:
        """Returns true if PGN matches DM1 (active DTC) PGN."""
        return self.getPGN() == 0xFECA
    
    def isDM2(self) -> bool:
        """Returns true if PGN matches DM2 (previously active DTC) PGN."""
        return self.getPGN() == 0xFECB
    
    def isDM3(self) -> bool:
        """Returns true if PGN matches DM3 (clear previously active DTCs) PGN."""
        return self.getPGN() == 0xFECC

    def isDM4(self) -> bool:
        """Returns true if PGN matches DM4 (freeze frame parameters) PGN."""
        return self.getPGN() == 0xFECD

    def isDM11(self) -> bool:
        """Returns true if PGN matches DM11 (clear active DTCs) PGN."""
        return self.getPGN() == 0xFED3

    def isDM12(self) -> bool:
        """Returns true if PGN matches DM12 (emission-related active DTCs) PGN."""
        return self.getPGN() == 0xFED4

class DiagnosticMessage():
    """
    A convenience class for parsing Diagnostic Messages DM1, DM2, and DM12.
    
    msg param types:
    - `J1939Message` - copies data directly from message data
    - `bytes` - data taken directly from RP1210_ReadMessage
    - `int` - data from J1939 message (w/o PGN, etc)
    """
    def __init__(self, msg = None) -> None:
        if msg is None:
            self.data = b'\x00\x00'
        elif isinstance(msg, J1939Message):
            self.data = msg.getData()
        elif isinstance(msg, bytes):
            self.data = J1939Message(msg).getData()
        else:
            self.data = sanitize_msg_param(msg)

    #############
    # FUNCTIONS #
    #############

    def mil(self):
        """MIL (malfunction indicator lamp) status (0-3)."""
        return (self.lamps[0] & 0b11000000) >> 6
    
    def rsl(self):
        """RSL (red stop lamp) status (0-3)."""
        return (self.lamps[0] & 0b00110000) >> 4

    def awl(self):
        """AWL (amber warning lamp) status (0-3)."""
        return (self.lamps[0] & 0b00001100) >> 2

    def pl(self):
        """PL (protection lamp) status (0-3)."""
        return (self.lamps[0] &0b00000011)

    ##################
    # DUNDER METHODS #
    ##################
    #region dundermethods

    def __getitem__(self, index : int) -> DTC:
        # O^2 since to_dtcs() generates a new array each time
        # generally relatively small, but probably worth optimizing anyway
        return self.to_dtcs(self.data)[index]
    
    def __setitem__(self, index : int, dtc : DTC):
        # like getitem, this is quite unoptimized
        data = b''
        # add lamps
        data += bytes(self.data[0])
        data += bytes(self.data[1])
        # copy old data into new data until we hit index
        for i in range(2, index * 4 + 2):
            data += bytes(self.data[i])
        # copy new dtc into data
        dtc = sanitize_msg_param(dtc)
        for i in range(0, 4):
            data += bytes(dtc[i])
        # copy rest of old data
        for i in range(index+4, len(self.data)):
            data += bytes(self.data[i])
        # set self.data = new data
        self.data = data

    def __iadd__(self, dtc : DTC):
        """Add DTCs to DiagnosticMessage."""
        self.data += sanitize_msg_param(dtc, 4)
        return self

    def __bytes__(self) -> bytes:
        return self.data

    def __int__(self) -> int:
        return int.from_bytes(self.data, 'big')

    def __str__(self) -> str:
        """Returns string representation of data."""
        return str(self.data)

    def __len__(self) -> int:
        """Returns number of DTCs stored in DiagnosticMessage object."""
        dtc_bytes = len(self.data) - 2
        num_dtcs = int(dtc_bytes / 4)
        return max(num_dtcs, 0)

    def __bool__(self) -> bool:
        return len(self.data) >= 6

    def __eq__(self, other) -> bool:
        try:
            return sanitize_msg_param(other) == sanitize_msg_param(self.data)
        except Exception:
            return False


    #endregion

    ##############
    # PROPERTIES #
    ##############
    #region properties

    @property
    def codes(self) -> list[DTC]:
        """List of DTC objects parsed from DiagnosticMessage."""
        return self.to_dtcs(self.data)

    @codes.setter
    def codes(self, val : bytes):
        """List of DTC objects parsed from DiagnosticMessage."""
        self.data = self.data[0:2] + val

    @property
    def lamps(self) -> bytes:
        """Byte 0 is lamp codes; Byte 1 is reserved."""
        if len(self.data) >= 2:
            return self.data[0:2]
        elif len(self.data) == 1:
            return int.to_bytes(self.data[0], 2, 'little')
        else:
            return b'\x00\x00'

    @lamps.setter
    def lamps(self, val : bytes):
        """Byte 0 is lamp codes; Byte 1 is reserved."""
        if isinstance(val, int) or len(val) == 1:
            lamp_code = sanitize_msg_param(val, 2, 'little')
        else:
            lamp_code = sanitize_msg_param(val, 2)
        self.data = lamp_code + self.data[2:]

    #endregion

    ########################
    # STATIC/CLASS METHODS #
    ########################
    #region staticmethods
    @staticmethod
    def to_dtcs(data : bytes) -> list[DTC]:
        """
        Parses given J1939 message data into a list of DTCs (diagnostic trouble codes).
        """
        dtcs = [] #type: list[DTC]
        for i in range(2, len(data), 4): # iterate in chunks of 4 bytes
            dtc = data[i:i+4]
            if len(dtc) == 4:
                dtcs.append(DTC(dtc))
        return dtcs

    #endregion

############################
# MOSTLY USELESS FUNCTIONS #
############################

def toJ1939Name(arbitrary_address : bool, industry_group : int, system_instance : int, system : int,
                function : int, function_instance : int, ecu_instance : int, mfg_code : int, id : int) -> bytes:
    """
    Each J1939-compliant ECU needs its own 64-bit name. This function is meant to help generate such
    a name based on the component bytes that make it up.
    """
    def add_bits(name, val, num_bits):
        mask = (1 << num_bits) - 1
        value = int.from_bytes(sanitize_msg_param(val), (num_bits + 7) // 8)
        return (name << num_bits) + (value & mask)
    name = 0b0
    # arbitrary address (1 bit)
    name = add_bits(name, arbitrary_address, 1)
    # industry group (3 bits)
    name = add_bits(name, industry_group, 3)
    # vehicle system instance (4 bits)
    name = add_bits(name, system_instance, 4)
    # vehicle system (7 bits)
    name = add_bits(name, system, 7)
    # reserved bit
    name = add_bits(name, 1, 1)
    # function (8 bits)
    name = add_bits(name, function, 8)
    # function instance (5 bits)
    name = add_bits(name, function_instance, 5)
    # ecu instance (3 bits)
    name = add_bits(name, ecu_instance, 3)
    # manufacturer code (11 bits)
    name = add_bits(name, mfg_code, 11)
    # identity number (21 bits)
    name = add_bits(name, id, 21)
    return sanitize_msg_param(name)

def getJ1939ProtocolString(protocol = 1, Baud = "Auto", Channel = -1,
                        SampleLocation = 95, SJW = 1,
                        PROP_SEG = 1, PHASE_SEG1 = 2, PHASE_SEG2 = 1,
                        TSEG1 = 2, TSEG2 = 1, SampleTimes = 1) -> bytes:
    """
    Generates fpchProtocol string for ClientConnect function. The default values you see above
    were made up on the spot and shouldn't necessarily be used.

    Keyword arguments have the same names as below (e.g. Baud, SampleLocation, PHASE_SEG1).

    IDSize is automatically set to 29 whenever relevant because that is its only valid value.

    This function also accepts a Channel argument!

    Examples:
    - protocol1 = J1939.getProtocolString(protocol = 1, Baud = "Auto")
    - protocol2 = J1939.getProtocolString(protocol = 3, Baud = 500, SampleLocation = 75, SJW = 3, IDSize = 29)
    """
    if Channel != -1:
        chan_arg = f",Channel={str(Channel)}"
    else:
        chan_arg = ""

    if protocol == 1:
        return bytes(f"J1939:Baud={str(Baud)}" + chan_arg, 'utf-8')
    elif protocol == 2:
        return bytes("J1939" + chan_arg, 'utf-8')
    elif protocol == 3:
        return bytes(f"J1939:Baud={str(Baud)},SampleLocation={str(SampleLocation)},SJW={str(SJW)},IDSize=29" + chan_arg, 'utf-8')
    elif protocol == 4:
        return bytes(f"J1939:Baud={str(Baud)},PROP_SEG={str(PROP_SEG)},PHASE_SEG1={str(PHASE_SEG1)},PHASE_SEG2={str(PHASE_SEG2)},SJW={str(SJW)},IDSize=29" + chan_arg, 'utf-8')
    elif protocol == 5:
        return bytes(f"J1939:Baud={str(Baud)},TSEG1={str(TSEG1)},TSEG2={str(TSEG2)},SampleTimes={str(SampleTimes)},SJW={str(SJW)},IDSize=29" + chan_arg, 'utf-8')
    else:
        return b"J1939" # default to protocol format 2, default channel

def getJ1939ProtocolDescription(protocol : int) -> str:
    """
    Returns a description of the protocol selected with protocol arg.
    
    Feed the result of RP1210Config.getJ1939FormatsSupported() into this function to get a description of what
    the format means.

    Honestly, I don't see anyone ever using this function.
    """
    if protocol == 1:
        return "Variable J1939 baud rate. Select 125, 250, 500, 1000, or Auto."
    elif protocol == 2:
        return "General default for J1939 baud rate (250k baud)."
    elif protocol == 3:
        return "Driver uses SampleLocation to calculate parameters."
    elif protocol == 4:
        return "Baud formula derived from BOSCH CAN specification."
    elif protocol == 5:
        return "Baud formula derived from Intel implementations."
    else:
        return "Invalid J1939 protocol format selected."
